@@ -33,7 +33,7 @@ int main(int argc, char **argv)
 	}
 
 	if ((fd = getfd(NULL)) < 0)
-		kbd_error(EXIT_FAILURE, 0, _("Couldn't get a file descriptor referring to the console"));
+		kbd_error(EXIT_FAILURE, 0, _("Couldn't get a file descriptor referring to the console."));
 
 	if (ioctl(fd, GIO_CMAP, cmap))
 		kbd_error(EXIT_FAILURE, errno, "ioctl GIO_CMAP");
